//
//	Copyright (C) by S.Tomabechi 1997
//	All rights reserved.
//
//	Primality test by method of Proth
//
//	<Proth's Theorem>
//	Let N = k*2^n + 1 with 2^n > k. If there is an integer a
//	such that a^((N-1)/2) = -1 (mod N), then N is prime.
//
//	return value   1   : prime
//				   0   : not prime
//				  -1   : cannot be tested by this routine
//				   m>0 : prime and a factor of Fermat number Fm
//
//	  cf Ribenboim	The Little Book of Big Primes	Springer Verlag
//		 D. Knuth  The Art of Computer Programming Volume II
//					 SemiNumerical Algorithms  problem 5.4.27
//

#include <stdlib.h>
#include "Mint.hpp"
#include "MBit.hpp"
#include "MintFunc.hpp"
#include "Error.hpp"

#define isPrime 1
#define notPrime 0
#define cannotTested -1

int Proth(const Mint& N)
{
  unsigned long d ;


  Mint N_1(N) ;
  --N_1 ;
  long n,m ;
  long k ;
  Mint p2 ;

  int j,l ;
  Mint a ;
  Mint A,B,r,q ;
  Mint h ;

  n=Val2(N_1) ;				  // max power of 2 which divides N-1
  p2=1 ;
  p2<<=n ;					  // p2=2^n
  --p2 ;
  h=(N_1>>n) ;				  // N=h*2^n+1
  N_1>>=1 ;					  //  (N-1)/2

							  // for debug
  if( (h*(p2+1)+1)!=N ) Error(decomposition error) ;
  if( p2<=h ) {
	cout << endl << "number cannot be tested by this routine " ;
	return cannotTested ;
  }

  d=prmdiv(N) ;	  // find small divisor
	 // N has a small divisor
  if( d>1 ) return notPrime ;

  long leng=ilog2(N_1) ;	  // bit lenght of (N-1)/2

  Mint a2,x ;

  for( a=2 ; a<N ; ++a ) {
	if( Kronecker(a,N)!=-1 ) continue ;	 // not quadratic residue ?
	a2=a ;
	x=1 ;
	l=leng ;
	if( l>=0 ) for( j=0 ;;){
	  d=N_1.digit[j++] ;

			  // x=A*2^n+B
			  // A=qh+r	   0<=r<h
			  // x=(qh+r)*2^n+B =B-q+r*2^n+ q*(h*2^n+1)
			  //  =B-q+r*2^n   mod h*2^n+1
	  for( k=32 ; k!=0 ; k-- ) {
		if( d&1 ) {
		  x*=a2 ;
//			A=(x>>n) ;		   // divide by 2^n
		  A.isShiftr(x,n) ; 
//			bB=bx&bp2 ;		 // mod 2^n
		  B.isRem2(x,n) ;	 // B=x mod 2^n
		  divide(A,h,q,r) ;
		  x=B+(r<<n)-q ;   // x mod N
		}

		if( --l<0 ) goto Check ;

		d>>=1 ;
		a2=square(a2)  ;
//		  A=(a2>>n) ;			  // a^2=A*2^n+B
		A.isShiftr(a2,n) ;
//		  bB=ba2&bp2 ;			  // B = a^2 mod 2^n
		B.isRem2(a2,n) ;
		divide(A,h,q,r) ;
		a2=B+(r<<n)-q ;			// a2 mod N
	  }
	}
Check: ;

	if( sign(x)>0 ) x-=N ;
	if( x==-1 ) goto Found ;
	else return notPrime ;
  }
  return notPrime ;
Found: ;
cout << endl << "divide Fermat number ? m=" ;

  a2=4 ;
  for( m=2 ; m<=n-2 ; m++ ) {
cout << " " << m  << flush ;
	a2=square(a2) ;
	A=(a2>>n) ;				// a^2=A*2^n+B
//	  bB=ba2&bp2 ;			  // B=a^2 mod 2^n
	B.isRem2(a2,n) ;
	divide(A,h,q,r) ;		// A=qh+r
	a2=B+(r<<n)-q ;			// a2 mod N
	if( sign(a2)>0 ) a2-=N ;
	if( a2==-1 ) return m ;
  }
  return isPrime ;
}

/**********
int main(void)
{
  Mint N ;
  long h=100  ;
  long k ;
  for( k=24 ; k<=1200 ; k++ ) {
	N=1 ;
	N<<=k ;
	for( h=1 ; h<N ; h+=2 ) {
cout << endl << h << "*" << "2^" << k << "+1" ;
	  if( Proth(h*N+1)<=0 ) {
		if( power(Mint(3),(h*N),h*N+1)==1 ) Error(this is prime) ;
		continue ;
	  }
	  if( !Rabin(h*N+1) ) Error( error) ;
	  cout << endl << " prime " << h << " " << k ;
	  break ;
	}
  }
  return 0 ;
}
***********/

